#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from eventlet import tpool
import mock
from oslo_concurrency import processutils
from oslo_serialization import jsonutils
from oslo_utils.fixture import uuidsentinel as uuids

from nova.compute import task_states
from nova import exception
from nova import objects
from nova.storage import rbd_utils
from nova import test


CEPH_MON_DUMP = r"""dumped monmap epoch 1
{ "epoch": 1,
  "fsid": "33630410-6d93-4d66-8e42-3b953cf194aa",
  "modified": "2013-05-22 17:44:56.343618",
  "created": "2013-05-22 17:44:56.343618",
  "mons": [
        { "rank": 0,
          "name": "a",
          "addr": "[::1]:6789\/0"},
        { "rank": 1,
          "name": "b",
          "addr": "[::1]:6790\/0"},
        { "rank": 2,
          "name": "c",
          "addr": "[::1]:6791\/0"},
        { "rank": 3,
          "name": "d",
          "addr": "127.0.0.1:6792\/0"},
        { "rank": 4,
          "name": "e",
          "addr": "example.com:6791\/0"}],
  "quorum": [
        0,
        1,
        2]}
"""


# max_avail stats are tweaked for testing
CEPH_DF = """
{
    "stats": {
        "total_bytes": 25757220864,
        "total_used_bytes": 274190336,
        "total_avail_bytes": 25483030528
    },
    "pools": [
        {
            "name": "images",
            "id": 1,
            "stats": {
                "kb_used": 12419,
                "bytes_used": 12716067,
                "percent_used": 0.05,
                "max_avail": 24195168123,
                "objects": 6
            }
        },
        {
            "name": "rbd",
            "id": 2,
            "stats": {
                "kb_used": 0,
                "bytes_used": 0,
                "percent_used": 0.00,
                "max_avail": 24195168456,
                "objects": 0
            }
        },
        {
            "name": "volumes",
            "id": 3,
            "stats": {
                "kb_used": 0,
                "bytes_used": 0,
                "percent_used": 0.00,
                "max_avail": 24195168789,
                "objects": 0
            }
        }
    ]
}
"""


class FakeException(Exception):
    pass


class RbdTestCase(test.NoDBTestCase):

    def setUp(self):
        super(RbdTestCase, self).setUp()

        self.rbd_pool = 'rbd'
        self.rbd_connect_timeout = 5
        self.flags(
            images_rbd_pool=self.rbd_pool,
            images_rbd_ceph_conf='/foo/bar.conf',
            rbd_connect_timeout=self.rbd_connect_timeout,
            rbd_user='foo', group='libvirt')

        rados_patcher = mock.patch.object(rbd_utils, 'rados')
        self.mock_rados = rados_patcher.start()
        self.addCleanup(rados_patcher.stop)
        self.mock_rados.Rados = mock.Mock()
        self.rados_inst = mock.Mock()
        self.mock_rados.Rados.return_value = self.rados_inst
        self.rados_inst.ioctx = mock.Mock()
        self.rados_inst.connect = mock.Mock()
        self.rados_inst.shutdown = mock.Mock()
        self.rados_inst.open_ioctx = mock.Mock()
        self.rados_inst.open_ioctx.return_value = \
            self.rados_inst.ioctx
        self.mock_rados.Error = Exception

        rbd_patcher = mock.patch.object(rbd_utils, 'rbd')
        self.mock_rbd = rbd_patcher.start()
        self.addCleanup(rbd_patcher.stop)
        self.mock_rbd.RBD = mock.Mock()
        self.mock_rbd.Image = mock.Mock()
        self.mock_rbd.Image.close = mock.Mock()
        self.mock_rbd.Error = Exception
        self.mock_rbd.ImageBusy = FakeException
        self.mock_rbd.ImageHasSnapshots = FakeException

        self.driver = rbd_utils.RBDDriver()

        self.volume_name = u'volume-00000001'
        self.snap_name = u'test-snap'

    def test_rbdproxy_wraps_rbd(self):
        proxy = rbd_utils.RbdProxy()
        self.assertIsInstance(proxy._rbd, tpool.Proxy)

    def test_rbdproxy_attribute_access_proxying(self):
        client = mock.MagicMock(ioctx='fake_ioctx')
        rbd_utils.RbdProxy().list(client.ioctx)
        self.mock_rbd.RBD.return_value.list.assert_called_once_with(
            client.ioctx)

    def test_good_locations(self):
        locations = ['rbd://fsid/pool/image/snap',
                     'rbd://%2F/%2F/%2F/%2F', ]
        map(self.driver.parse_url, locations)

    def test_bad_locations(self):
        locations = ['rbd://image',
                     'http://path/to/somewhere/else',
                     'rbd://image/extra',
                     'rbd://image/',
                     'rbd://fsid/pool/image/',
                     'rbd://fsid/pool/image/snap/',
                     'rbd://///', ]
        image_meta = {'disk_format': 'raw'}

        for loc in locations:
            self.assertRaises(exception.ImageUnacceptable,
                              self.driver.parse_url, loc)
            self.assertFalse(self.driver.is_cloneable({'url': loc},
                                                      image_meta))

    @mock.patch.object(rbd_utils, 'RADOSClient')
    def test_rbddriver(self, mock_client):
        client = mock_client.return_value
        client.__enter__.return_value = client
        client.cluster.get_fsid.side_effect = lambda: b'abc'
        self.assertEqual('abc', self.driver.get_fsid())

    @mock.patch.object(rbd_utils.RBDDriver, 'get_fsid')
    def test_cloneable(self, mock_get_fsid):
        mock_get_fsid.return_value = 'abc'
        location = {'url': 'rbd://abc/pool/image/snap'}
        image_meta = {'disk_format': 'raw'}
        self.assertTrue(self.driver.is_cloneable(location, image_meta))
        self.assertTrue(mock_get_fsid.called)

    @mock.patch.object(rbd_utils.RBDDriver, 'get_fsid')
    def test_uncloneable_different_fsid(self, mock_get_fsid):
        mock_get_fsid.return_value = 'abc'
        location = {'url': 'rbd://def/pool/image/snap'}
        image_meta = {'disk_format': 'raw'}
        self.assertFalse(
            self.driver.is_cloneable(location, image_meta))
        self.assertTrue(mock_get_fsid.called)

    @mock.patch.object(rbd_utils.RBDDriver, 'get_fsid')
    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_uncloneable_unreadable(self, mock_proxy,
                                    mock_get_fsid):
        mock_get_fsid.return_value = 'abc'
        location = {'url': 'rbd://abc/pool/image/snap'}

        mock_proxy.side_effect = self.mock_rbd.Error
        image_meta = {'disk_format': 'raw'}

        self.assertFalse(
            self.driver.is_cloneable(location, image_meta))
        mock_proxy.assert_called_once_with(self.driver, 'image', pool='pool',
                                           snapshot='snap', read_only=True)
        self.assertTrue(mock_get_fsid.called)

    @mock.patch.object(rbd_utils.RBDDriver, 'get_fsid')
    def test_uncloneable_bad_format(self, mock_get_fsid):
        mock_get_fsid.return_value = 'abc'
        location = {'url': 'rbd://abc/pool/image/snap'}
        formats = ['qcow2', 'vmdk', 'vdi']
        for f in formats:
            image_meta = {'disk_format': f}
            self.assertFalse(
                self.driver.is_cloneable(location, image_meta))
        self.assertTrue(mock_get_fsid.called)

    @mock.patch.object(rbd_utils.RBDDriver, 'get_fsid')
    def test_uncloneable_missing_format(self, mock_get_fsid):
        mock_get_fsid.return_value = 'abc'
        location = {'url': 'rbd://abc/pool/image/snap'}
        image_meta = {}
        self.assertFalse(
            self.driver.is_cloneable(location, image_meta))
        self.assertTrue(mock_get_fsid.called)

    @mock.patch('oslo_concurrency.processutils.execute')
    def test_get_mon_addrs(self, mock_execute):
        mock_execute.return_value = (CEPH_MON_DUMP, '')
        hosts = ['::1', '::1', '::1', '127.0.0.1', 'example.com']
        ports = ['6789', '6790', '6791', '6792', '6791']
        self.assertEqual((hosts, ports), self.driver.get_mon_addrs())

    @mock.patch('oslo_concurrency.processutils.execute')
    def test_get_mon_addrs_with_brackets(self, mock_execute):
        mock_execute.return_value = (CEPH_MON_DUMP, '')
        hosts = ['[::1]', '[::1]', '[::1]', '127.0.0.1', 'example.com']
        ports = ['6789', '6790', '6791', '6792', '6791']
        self.assertEqual((hosts, ports),
                         self.driver.get_mon_addrs(strip_brackets=False))

    @mock.patch.object(rbd_utils.RBDDriver, '_connect_to_rados')
    def test_rbd_conf_features(self, mock_connect):
        self.mock_rbd.RBD_FEATURE_LAYERING = 1
        mock_cluster = mock.Mock()
        mock_cluster.conf_get = mock.Mock()
        mock_cluster.conf_get.return_value = None
        mock_connect.return_value = (mock_cluster, None)
        client = rbd_utils.RADOSClient(self.driver)
        self.assertEqual(1, client.features)

        mock_cluster.conf_get.return_value = '2'
        self.assertEqual(2, client.features)

    @mock.patch.object(rbd_utils, 'RADOSClient')
    def test_clone(self, mock_client):
        pool = u'images'
        image = u'image-name'
        snap = u'snapshot-name'
        location = {'url': u'rbd://fsid/%s/%s/%s' % (pool, image, snap)}

        client_stack = []

        def mock__enter__(inst):
            def _inner():
                client_stack.append(inst)
                return inst
            return _inner

        client = mock_client.return_value
        # capture both rados client used to perform the clone
        client.__enter__.side_effect = mock__enter__(client)

        rbd = self.mock_rbd.RBD.return_value

        self.driver.clone(location, self.volume_name)

        args = [client_stack[0].ioctx, image, snap,
                client_stack[1].ioctx, str(self.volume_name)]
        kwargs = {'features': client.features}
        rbd.clone.assert_called_once_with(*args, **kwargs)
        self.assertEqual(2, client.__enter__.call_count)

    @mock.patch.object(rbd_utils, 'RADOSClient')
    def test_clone_eperm(self, mock_client):
        pool = u'images'
        image = u'image-name'
        snap = u'snapshot-name'
        location = {'url': u'rbd://fsid/%s/%s/%s' % (pool, image, snap)}

        client_stack = []

        def mock__enter__(inst):
            def _inner():
                client_stack.append(inst)
                return inst
            return _inner

        client = mock_client.return_value
        # capture both rados client used to perform the clone
        client.__enter__.side_effect = mock__enter__(client)

        setattr(self.mock_rbd, 'PermissionError', test.TestingException)
        rbd = self.mock_rbd.RBD.return_value
        rbd.clone.side_effect = test.TestingException
        self.assertRaises(exception.Forbidden,
                          self.driver.clone, location, self.volume_name)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_resize(self, mock_proxy):
        size = 1024
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        self.driver.resize(self.volume_name, size)
        proxy.resize.assert_called_once_with(size)

    @mock.patch.object(rbd_utils.RBDDriver, '_disconnect_from_rados')
    @mock.patch.object(rbd_utils.RBDDriver, '_connect_to_rados')
    def test_rbd_volume_proxy_init(self, mock_connect_from_rados,
                                   mock_disconnect_from_rados):
        mock_connect_from_rados.return_value = (None, None)
        mock_disconnect_from_rados.return_value = (None, None)

        with rbd_utils.RBDVolumeProxy(self.driver, self.volume_name):
            mock_connect_from_rados.assert_called_once_with(None)
            self.assertFalse(mock_disconnect_from_rados.called)

        mock_disconnect_from_rados.assert_called_once_with(None, None)

    def test_connect_to_rados_default(self):
        ret = self.driver._connect_to_rados()
        self.rados_inst.connect.assert_called_once_with(
                timeout=self.rbd_connect_timeout)
        self.assertTrue(self.rados_inst.open_ioctx.called)
        self.assertEqual(self.rados_inst.ioctx, ret[1])
        self.rados_inst.open_ioctx.assert_called_with(self.rbd_pool)

    def test_connect_to_rados_different_pool(self):
        ret = self.driver._connect_to_rados('alt_pool')
        self.rados_inst.connect.assert_called_once_with(
                timeout=self.rbd_connect_timeout)
        self.assertTrue(self.rados_inst.open_ioctx.called)
        self.assertEqual(self.rados_inst.ioctx, ret[1])
        self.rados_inst.open_ioctx.assert_called_with('alt_pool')

    def test_connect_to_rados_error(self):
        self.rados_inst.open_ioctx.side_effect = self.mock_rados.Error
        self.assertRaises(self.mock_rados.Error,
                          self.driver._connect_to_rados)
        self.rados_inst.open_ioctx.assert_called_once_with(
            self.rbd_pool)
        self.rados_inst.shutdown.assert_called_once_with()

    def test_connect_to_rados_unicode_arg(self):
        self.driver._connect_to_rados(u'unicode_pool')
        self.rados_inst.open_ioctx.assert_called_with(
            test.MatchType(str))

    def test_ceph_args_none(self):
        self.driver.rbd_user = None
        self.driver.ceph_conf = None
        self.assertEqual([], self.driver.ceph_args())

    def test_ceph_args_rbd_user(self):
        self.driver.rbd_user = 'foo'
        self.driver.ceph_conf = None
        self.assertEqual(['--id', 'foo'], self.driver.ceph_args())

    def test_ceph_args_ceph_conf(self):
        self.driver.rbd_user = None
        self.driver.ceph_conf = '/path/bar.conf'
        self.assertEqual(['--conf', '/path/bar.conf'],
                         self.driver.ceph_args())

    def test_ceph_args_rbd_user_and_ceph_conf(self):
        self.driver.rbd_user = 'foo'
        self.driver.ceph_conf = '/path/bar.conf'
        self.assertEqual(['--id', 'foo', '--conf', '/path/bar.conf'],
                         self.driver.ceph_args())

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_exists(self, mock_proxy):
        snapshot = 'snap'
        proxy = mock_proxy.return_value
        self.assertTrue(self.driver.exists(self.volume_name,
                                           self.rbd_pool,
                                           snapshot))
        proxy.__enter__.assert_called_once_with()
        proxy.__exit__.assert_called_once_with(None, None, None)

    @mock.patch.object(rbd_utils, 'RADOSClient')
    def test_cleanup_volumes(self, mock_client):
        instance = objects.Instance(id=1, uuid=uuids.instance,
                                    task_state=None)
        # this is duplicated from nova/virt/libvirt/driver.py
        filter_fn = lambda disk: disk.startswith(instance.uuid)

        rbd = self.mock_rbd.RBD.return_value
        rbd.list.return_value = ['%s_test' % uuids.instance, '111_test']

        client = mock_client.return_value
        self.driver.cleanup_volumes(filter_fn)

        rbd.remove.assert_called_once_with(
            client.__enter__.return_value.ioctx,
            '%s_test' % uuids.instance)
        client.__enter__.assert_called_once_with()
        client.__exit__.assert_called_once_with(None, None, None)

    @mock.patch('oslo_service.loopingcall.LoopingCallBase._sleep',
                new=mock.Mock())
    @mock.patch.object(rbd_utils, 'RADOSClient')
    def _test_cleanup_exception(self, exception_name, mock_client):
        instance = objects.Instance(id=1, uuid=uuids.instance,
                                    task_state=None)
        # this is duplicated from nova/virt/libvirt/driver.py
        filter_fn = lambda disk: disk.startswith(instance.uuid)

        setattr(self.mock_rbd, exception_name, test.TestingException)
        rbd = self.mock_rbd.RBD.return_value
        rbd.remove.side_effect = test.TestingException
        rbd.list.return_value = ['%s_test' % uuids.instance, '111_test']
        self.mock_rbd.Image.return_value.list_snaps.return_value = [{}]

        client = mock_client.return_value
        self.driver.cleanup_volumes(filter_fn)
        rbd.remove.assert_any_call(client.__enter__.return_value.ioctx,
                                   '%s_test' % uuids.instance)
        # NOTE(sandonov): 12 retries + 1 final attempt to propagate = 13
        self.assertEqual(13, len(rbd.remove.call_args_list))

    def test_cleanup_volumes_fail_not_found(self):
        self._test_cleanup_exception('ImageBusy')

    def test_cleanup_volumes_fail_snapshots(self):
        self._test_cleanup_exception('ImageHasSnapshots')

    def test_cleanup_volumes_fail_other(self):
        self.assertRaises(test.TestingException,
                          self._test_cleanup_exception, 'DoesNotExist')

    @mock.patch.object(rbd_utils, 'RADOSClient')
    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_cleanup_volumes_pending_resize(self, mock_proxy, mock_client):
        self.mock_rbd.ImageBusy = FakeException
        self.mock_rbd.ImageHasSnapshots = FakeException
        instance = objects.Instance(id=1, uuid=uuids.instance,
                                    task_state=None)
        # this is duplicated from nova/virt/libvirt/driver.py
        filter_fn = lambda disk: disk.startswith(instance.uuid)

        setattr(self.mock_rbd, 'ImageHasSnapshots', test.TestingException)
        rbd = self.mock_rbd.RBD.return_value
        rbd.remove.side_effect = [test.TestingException, None]
        rbd.list.return_value = ['%s_test' % uuids.instance, '111_test']
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        proxy.list_snaps.return_value = [
            {'name': rbd_utils.RESIZE_SNAPSHOT_NAME}]
        client = mock_client.return_value
        self.driver.cleanup_volumes(filter_fn)

        remove_call = mock.call(client.__enter__.return_value.ioctx,
                                '%s_test' % uuids.instance)
        rbd.remove.assert_has_calls([remove_call, remove_call])
        proxy.remove_snap.assert_called_once_with(
            rbd_utils.RESIZE_SNAPSHOT_NAME)
        client.__enter__.assert_called_once_with()
        client.__exit__.assert_called_once_with(None, None, None)

    @mock.patch.object(rbd_utils, 'RADOSClient')
    def test_cleanup_volumes_reverting_resize(self, mock_client):
        instance = objects.Instance(id=1, uuid=uuids.instance,
                                    task_state=task_states.RESIZE_REVERTING)
        # this is duplicated from nova/virt/libvirt/driver.py
        filter_fn = lambda disk: (disk.startswith(instance.uuid) and
                                  disk.endswith('disk.local'))

        rbd = self.mock_rbd.RBD.return_value
        rbd.list.return_value = ['%s_test' % uuids.instance, '111_test',
                                 '%s_test_disk.local' % uuids.instance]

        client = mock_client.return_value
        self.driver.cleanup_volumes(filter_fn)
        rbd.remove.assert_called_once_with(
            client.__enter__.return_value.ioctx,
            '%s_test_disk.local' % uuids.instance)
        client.__enter__.assert_called_once_with()
        client.__exit__.assert_called_once_with(None, None, None)

    @mock.patch.object(rbd_utils, 'RADOSClient')
    def test_destroy_volume(self, mock_client):
        rbd = self.mock_rbd.RBD.return_value
        vol = '12345_test'
        client = mock_client.return_value
        self.driver.destroy_volume(vol)
        rbd.remove.assert_called_once_with(
            client.__enter__.return_value.ioctx, vol)
        client.__enter__.assert_called_once_with()
        client.__exit__.assert_called_once_with(None, None, None)

    @mock.patch.object(rbd_utils, 'RADOSClient')
    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall')
    def test_destroy_volume_with_retries(self, mock_loopingcall, mock_client):
        vol = '12345_test'
        client = mock_client.return_value
        loopingcall = mock_loopingcall.return_value

        # Try for sixty seconds: six retries at 10 second interval
        self.flags(rbd_destroy_volume_retries=6, group='libvirt')
        self.flags(rbd_destroy_volume_retry_interval=10, group='libvirt')
        self.driver.destroy_volume(vol)

        # Make sure both params have the expected values
        retryctx = mock_loopingcall.call_args.args[3]
        self.assertEqual(retryctx, {'retries': 6})
        loopingcall.start.assert_called_with(interval=10)

        # Make sure that we entered and exited the RADOSClient
        client.__enter__.assert_called_once_with()
        client.__exit__.assert_called_once_with(None, None, None)

    @mock.patch.object(rbd_utils, 'RADOSClient')
    def test_remove_image(self, mock_client):
        name = '12345_disk.config.rescue'

        rbd = self.mock_rbd.RBD.return_value

        client = mock_client.return_value
        self.driver.remove_image(name)
        rbd.remove.assert_called_once_with(
            client.__enter__.return_value.ioctx, name)
        # Make sure that we entered and exited the RADOSClient
        client.__enter__.assert_called_once_with()
        client.__exit__.assert_called_once_with(None, None, None)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_create_snap(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        self.driver.create_snap(self.volume_name, self.snap_name)
        proxy.create_snap.assert_called_once_with(self.snap_name)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_create_protected_snap(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        proxy.is_protected_snap.return_value = False
        self.driver.create_snap(self.volume_name, self.snap_name, protect=True)
        proxy.create_snap.assert_called_once_with(self.snap_name)
        proxy.is_protected_snap.assert_called_once_with(self.snap_name)
        proxy.protect_snap.assert_called_once_with(self.snap_name)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_remove_snap(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        proxy.list_snaps.return_value = [{'name': self.snap_name}]
        proxy.is_protected_snap.return_value = False
        self.driver.remove_snap(self.volume_name, self.snap_name)
        proxy.remove_snap.assert_called_once_with(self.snap_name)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_remove_snap_force(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        proxy.is_protected_snap.return_value = True
        proxy.list_snaps.return_value = [{'name': self.snap_name}]
        self.driver.remove_snap(self.volume_name, self.snap_name, force=True)
        proxy.is_protected_snap.assert_called_once_with(self.snap_name)
        proxy.unprotect_snap.assert_called_once_with(self.snap_name)
        proxy.remove_snap.assert_called_once_with(self.snap_name)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_remove_snap_does_nothing_when_no_snapshot(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        proxy.list_snaps.return_value = [{'name': 'some-other-snaphot'}]
        self.driver.remove_snap(self.volume_name, self.snap_name)
        self.assertFalse(proxy.remove_snap.called)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_remove_snap_does_nothing_when_protected(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        proxy.is_protected_snap.return_value = True
        proxy.list_snaps.return_value = [{'name': self.snap_name}]
        self.driver.remove_snap(self.volume_name, self.snap_name)
        self.assertFalse(proxy.remove_snap.called)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_remove_snap_protected_ignore_errors(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        proxy.is_protected_snap.return_value = True
        proxy.list_snaps.return_value = [{'name': self.snap_name}]
        self.driver.remove_snap(self.volume_name, self.snap_name,
                                ignore_errors=True)
        proxy.remove_snap.assert_called_once_with(self.snap_name)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_parent_info(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        self.driver.parent_info(self.volume_name)
        proxy.parent_info.assert_called_once_with()

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_parent_info_throws_exception_on_error(self, mock_proxy):
        setattr(self.mock_rbd, 'ImageNotFound', test.TestingException)
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        proxy.parent_info.side_effect = test.TestingException
        self.assertRaises(exception.ImageUnacceptable,
                          self.driver.parent_info, self.volume_name)

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_flatten(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        self.driver.flatten(self.volume_name)
        proxy.flatten.assert_called_once_with()

    @mock.patch.object(rbd_utils, 'RBDVolumeProxy')
    def test_rollback_to_snap(self, mock_proxy):
        proxy = mock_proxy.return_value
        proxy.__enter__.return_value = proxy
        self.assertRaises(exception.SnapshotNotFound,
                          self.driver.rollback_to_snap,
                          self.volume_name, self.snap_name)

        proxy.list_snaps.return_value = [{'name': self.snap_name}, ]
        self.driver.rollback_to_snap(self.volume_name, self.snap_name)
        proxy.rollback_to_snap.assert_called_once_with(self.snap_name)

    @mock.patch('oslo_concurrency.processutils.execute')
    def test_get_pool_info(self, mock_execute):
        mock_execute.return_value = (CEPH_DF, '')
        ceph_df_json = jsonutils.loads(CEPH_DF)
        expected = {'total': ceph_df_json['stats']['total_bytes'],
                    'free': ceph_df_json['pools'][1]['stats']['max_avail'],
                    'used': ceph_df_json['pools'][1]['stats']['bytes_used']}
        self.assertDictEqual(expected, self.driver.get_pool_info())

    @mock.patch('oslo_concurrency.processutils.execute', autospec=True,
                side_effect=processutils.ProcessExecutionError("failed"))
    def test_get_pool_info_execute_failed(self, mock_execute):
        self.assertRaises(exception.StorageError, self.driver.get_pool_info)

    @mock.patch('oslo_concurrency.processutils.execute')
    def test_get_pool_info_not_found(self, mock_execute):
        # Make the pool something other than self.rbd_pool so it won't be found
        ceph_df_not_found = CEPH_DF.replace('rbd', 'vms')
        mock_execute.return_value = (ceph_df_not_found, '')
        self.assertRaises(exception.NotFound, self.driver.get_pool_info)

    @mock.patch('oslo_concurrency.processutils.execute')
    def test_export_image(self, mock_execute):
        self.driver.rbd_user = 'foo'
        self.driver.export_image(mock.sentinel.dst_path,
                                 mock.sentinel.name,
                                 mock.sentinel.snap,
                                 mock.sentinel.pool)

        mock_execute.assert_called_once_with(
            'rbd', 'export',
            '--pool', mock.sentinel.pool,
            '--image', mock.sentinel.name,
            '--path', mock.sentinel.dst_path,
            '--snap', mock.sentinel.snap,
            '--id', 'foo',
            '--conf', '/foo/bar.conf')

    @mock.patch('oslo_concurrency.processutils.execute')
    def test_export_image_default_pool(self, mock_execute):
        self.driver.export_image(mock.sentinel.dst_path,
                                 mock.sentinel.name,
                                 mock.sentinel.snap)

        mock_execute.assert_called_once_with(
            'rbd', 'export',
            '--pool', self.rbd_pool,
            '--image', mock.sentinel.name,
            '--path', mock.sentinel.dst_path,
            '--snap', mock.sentinel.snap,
            '--id', 'foo',
            '--conf', '/foo/bar.conf')



# Copyright 2012 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Management class for VM snapshot operations.
"""
import os

from os_win import utilsfactory
from oslo_log import log as logging

from nova.compute import task_states
from nova.image import glance
from nova.virt.hyperv import pathutils

LOG = logging.getLogger(__name__)


class SnapshotOps(object):
    def __init__(self):
        self._pathutils = pathutils.PathUtils()
        self._vmutils = utilsfactory.get_vmutils()
        self._vhdutils = utilsfactory.get_vhdutils()

    def _save_glance_image(self, context, image_id, image_vhd_path):
        (glance_image_service,
         image_id) = glance.get_remote_image_service(context, image_id)
        image_metadata = {"disk_format": "vhd",
                          "container_format": "bare"}
        with self._pathutils.open(image_vhd_path, 'rb') as f:
            glance_image_service.update(context, image_id, image_metadata, f,
                                        purge_props=False)

    def snapshot(self, context, instance, image_id, update_task_state):
        """Create snapshot from a running VM instance."""
        instance_name = instance.name

        LOG.debug("Creating snapshot for instance %s", instance_name)
        snapshot_path = self._vmutils.take_vm_snapshot(instance_name)
        update_task_state(task_state=task_states.IMAGE_PENDING_UPLOAD)

        export_dir = None

        try:
            src_vhd_path = self._pathutils.lookup_root_vhd_path(instance_name)

            LOG.debug("Getting info for VHD %s", src_vhd_path)
            src_base_disk_path = self._vhdutils.get_vhd_parent_path(
                src_vhd_path)

            export_dir = self._pathutils.get_export_dir(instance_name)

            dest_vhd_path = os.path.join(export_dir, os.path.basename(
                src_vhd_path))
            LOG.debug('Copying VHD %(src_vhd_path)s to %(dest_vhd_path)s',
                      {'src_vhd_path': src_vhd_path,
                       'dest_vhd_path': dest_vhd_path})
            self._pathutils.copyfile(src_vhd_path, dest_vhd_path)

            image_vhd_path = None
            if not src_base_disk_path:
                image_vhd_path = dest_vhd_path
            else:
                basename = os.path.basename(src_base_disk_path)
                dest_base_disk_path = os.path.join(export_dir, basename)
                LOG.debug('Copying base disk %(src_vhd_path)s to '
                          '%(dest_base_disk_path)s',
                          {'src_vhd_path': src_vhd_path,
                           'dest_base_disk_path': dest_base_disk_path})
                self._pathutils.copyfile(src_base_disk_path,
                                         dest_base_disk_path)

                LOG.debug("Reconnecting copied base VHD "
                          "%(dest_base_disk_path)s and diff "
                          "VHD %(dest_vhd_path)s",
                          {'dest_base_disk_path': dest_base_disk_path,
                           'dest_vhd_path': dest_vhd_path})
                self._vhdutils.reconnect_parent_vhd(dest_vhd_path,
                                                    dest_base_disk_path)

                LOG.debug("Merging diff disk %s into its parent.",
                          dest_vhd_path)
                self._vhdutils.merge_vhd(dest_vhd_path)
                image_vhd_path = dest_base_disk_path

            LOG.debug("Updating Glance image %(image_id)s with content from "
                      "merged disk %(image_vhd_path)s",
                      {'image_id': image_id, 'image_vhd_path': image_vhd_path})
            update_task_state(task_state=task_states.IMAGE_UPLOADING,
                              expected_state=task_states.IMAGE_PENDING_UPLOAD)
            self._save_glance_image(context, image_id, image_vhd_path)

            LOG.debug("Snapshot image %(image_id)s updated for VM "
                      "%(instance_name)s",
                      {'image_id': image_id, 'instance_name': instance_name})
        finally:
            try:
                LOG.debug("Removing snapshot %s", image_id)
                self._vmutils.remove_vm_snapshot(snapshot_path)
            except Exception:
                LOG.exception('Failed to remove snapshot for VM %s',
                              instance_name, instance=instance)
            if export_dir:
                LOG.debug('Removing directory: %s', export_dir)
                self._pathutils.rmtree(export_dir)



# Copyright (c) 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Io Ops Weigher. Weigh hosts by their io ops number.

The default is to preferably choose light workload compute hosts. If you prefer
choosing heavy workload compute hosts, you can set 'io_ops_weight_multiplier'
option (by configuration or aggregate metadata) to a positive number and the
weighing has the opposite effect of the default.
"""

import nova.conf
from nova.scheduler import utils
from nova.scheduler import weights

CONF = nova.conf.CONF


class IoOpsWeigher(weights.BaseHostWeigher):
    minval = 0

    def weight_multiplier(self, host_state):
        """Override the weight multiplier."""
        return utils.get_weight_multiplier(
            host_state, 'io_ops_weight_multiplier',
            CONF.filter_scheduler.io_ops_weight_multiplier)

    def _weigh_object(self, host_state, weight_properties):
        """Higher weights win. We want to choose light workload host
        to be the default.
        """
        return host_state.num_io_ops



# Copyright 2015, 2018 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities related to glance image management for the PowerVM driver."""

from nova import utils


def stream_blockdev_to_glance(context, image_api, image_id, metadata, devpath):
    """Stream the entire contents of a block device to a glance image.

    :param context: Nova security context.
    :param image_api: Handle to the glance image API.
    :param image_id: UUID of the prepared glance image.
    :param metadata: Dictionary of metadata for the image.
    :param devpath: String path to device file of block device to be uploaded,
                    e.g. "/dev/sde".
    """
    # Make the device file owned by the current user for the duration of the
    # operation.
    with utils.temporary_chown(devpath), open(devpath, 'rb') as stream:
        # Stream it.  This is synchronous.
        image_api.update(context, image_id, metadata, stream)


def generate_snapshot_metadata(context, image_api, image_id, instance):
    """Generate a metadata dictionary for an instance snapshot.

    :param context: Nova security context.
    :param image_api: Handle to the glance image API.
    :param image_id: UUID of the prepared glance image.
    :param instance: The Nova instance whose disk is to be snapshotted.
    :return: A dict of metadata suitable for image_api.update.
    """
    image = image_api.get(context, image_id)

    # TODO(esberglu): Update this to v2 metadata
    metadata = {
        'name': image['name'],
        'status': 'active',
        'disk_format': 'raw',
        'container_format': 'bare',
        'properties': {
            'image_location': 'snapshot',
            'image_state': 'available',
            'owner_id': instance.project_id,
        }
    }
    return metadata



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional import fixtures as func_fixtures
from nova.tests.functional import integrated_helpers


class ColdMigrateTargetHostThenLiveMigrateTest(
        test.TestCase, integrated_helpers.InstanceHelperMixin):
    """Regression test for bug 1797580 introduced in Queens.

    Microversion 2.56 allows cold migrating to a specified target host. The
    compute API sets the requested destination on the request spec with the
    specified target host and then conductor sends that request spec to the
    scheduler to validate the host. Conductor later persists the changes to
    the request spec because it's the resize flow and the flavor could change
    (even though in this case it won't since it's a cold migrate). After
    confirming the resize, if the server is live migrated it will fail during
    scheduling because of the persisted RequestSpec.requested_destination
    from the cold migration, and you can't live migrate to the same host on
    which the instance is currently running.

    This test reproduces the regression and will validate the fix.
    """

    def setUp(self):
        super(ColdMigrateTargetHostThenLiveMigrateTest, self).setUp()
        self.useFixture(nova_fixtures.RealPolicyFixture())
        self.useFixture(nova_fixtures.NeutronFixture(self))
        self.glance = self.useFixture(nova_fixtures.GlanceFixture(self))
        self.useFixture(func_fixtures.PlacementFixture())

        api_fixture = self.useFixture(nova_fixtures.OSAPIFixture(
            api_version='v2.1'))
        # The admin API is used to get the server details to verify the
        # host on which the server was built and cold/live migrate it.
        self.admin_api = api_fixture.admin_api
        self.api = api_fixture.api
        # Use the latest microversion available to make sure something does
        # not regress in new microversions; cap as necessary.
        self.admin_api.microversion = 'latest'
        self.api.microversion = 'latest'

        self.start_service('conductor')
        self.start_service('scheduler')

        for host in ('host1', 'host2'):
            self.start_service('compute', host=host)

    def test_cold_migrate_target_host_then_live_migrate(self):
        # Create a server, it doesn't matter on which host it builds.
        server = self._build_server(networks='none')
        server = self.api.post_server({'server': server})
        server = self._wait_for_state_change(server, 'ACTIVE')
        original_host = server['OS-EXT-SRV-ATTR:host']
        target_host = 'host1' if original_host == 'host2' else 'host2'

        # Cold migrate the server to the specific target host.
        migrate_req = {'migrate': {'host': target_host}}
        self.admin_api.post_server_action(server['id'], migrate_req)
        server = self._wait_for_state_change(server, 'VERIFY_RESIZE')

        # Confirm the resize so the server stays on the target host.
        confim_req = {'confirmResize': None}
        self.admin_api.post_server_action(server['id'], confim_req)
        server = self._wait_for_state_change(server, 'ACTIVE')

        # Attempt to live migrate the server but don't specify a host so the
        # scheduler has to pick one.
        live_migrate_req = {
            'os-migrateLive': {'host': None, 'block_migration': 'auto'}}
        self.admin_api.post_server_action(server['id'], live_migrate_req)
        server = self._wait_for_state_change(server, 'ACTIVE')
        # The live migration should have been successful and the server is now
        # back on the original host.
        self.assertEqual(original_host, server['OS-EXT-SRV-ATTR:host'])



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.api.openstack import api_version_request
from nova import exception
from nova import test
from nova.tests.unit.api.openstack import fakes


class APIVersionRequestTests(test.NoDBTestCase):
    base_path = '/%s' % fakes.FAKE_PROJECT_ID

    def test_valid_version_strings(self):
        def _test_string(version, exp_major, exp_minor):
            v = api_version_request.APIVersionRequest(version)
            self.assertEqual(v.ver_major, exp_major)
            self.assertEqual(v.ver_minor, exp_minor)

        _test_string("1.1", 1, 1)
        _test_string("2.10", 2, 10)
        _test_string("5.234", 5, 234)
        _test_string("12.5", 12, 5)
        _test_string("2.0", 2, 0)
        _test_string("2.200", 2, 200)

    def test_null_version(self):
        v = api_version_request.APIVersionRequest()
        self.assertTrue(v.is_null())

    def test_invalid_version_strings(self):
        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "2")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "200")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "2.1.4")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "200.23.66.3")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "5 .3")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "5. 3")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "5.03")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "02.1")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "2.001")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, " 2.1")

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest, "2.1 ")

    def test_version_comparisons(self):
        vers1 = api_version_request.APIVersionRequest("2.0")
        vers2 = api_version_request.APIVersionRequest("2.5")
        vers3 = api_version_request.APIVersionRequest("5.23")
        vers4 = api_version_request.APIVersionRequest("2.0")
        v_null = api_version_request.APIVersionRequest()

        self.assertLess(v_null, vers2)
        self.assertLess(vers1, vers2)
        self.assertLessEqual(vers1, vers2)
        self.assertLessEqual(vers1, vers4)
        self.assertGreater(vers2, v_null)
        self.assertGreater(vers3, vers2)
        self.assertGreaterEqual(vers1, vers4)
        self.assertGreaterEqual(vers3, vers2)
        self.assertNotEqual(vers1, vers2)
        self.assertEqual(vers1, vers4)
        self.assertNotEqual(vers1, v_null)
        self.assertEqual(v_null, v_null)
        self.assertRaises(TypeError, vers1.__lt__, "2.1")

    def test_version_matches(self):
        vers1 = api_version_request.APIVersionRequest("2.0")
        vers2 = api_version_request.APIVersionRequest("2.5")
        vers3 = api_version_request.APIVersionRequest("2.45")
        vers4 = api_version_request.APIVersionRequest("3.3")
        vers5 = api_version_request.APIVersionRequest("3.23")
        vers6 = api_version_request.APIVersionRequest("2.0")
        vers7 = api_version_request.APIVersionRequest("3.3")
        vers8 = api_version_request.APIVersionRequest("4.0")
        v_null = api_version_request.APIVersionRequest()

        self.assertTrue(vers2.matches(vers1, vers3))
        self.assertTrue(vers2.matches(vers1, v_null))
        self.assertTrue(vers1.matches(vers6, vers2))
        self.assertTrue(vers4.matches(vers2, vers7))
        self.assertTrue(vers4.matches(v_null, vers7))
        self.assertTrue(vers4.matches(v_null, vers8))
        self.assertFalse(vers1.matches(vers2, vers3))
        self.assertFalse(vers5.matches(vers2, vers4))
        self.assertFalse(vers2.matches(vers3, vers1))

        self.assertRaises(ValueError, v_null.matches, vers1, vers3)

    def test_get_string(self):
        vers1_string = "3.23"
        vers1 = api_version_request.APIVersionRequest(vers1_string)
        self.assertEqual(vers1_string, vers1.get_string())

        self.assertRaises(ValueError,
                          api_version_request.APIVersionRequest().get_string)

    def test_is_supported_min_version(self):
        req = fakes.HTTPRequest.blank(self.base_path, version='2.5')

        self.assertTrue(api_version_request.is_supported(
            req, min_version='2.4'))
        self.assertTrue(api_version_request.is_supported(
            req, min_version='2.5'))
        self.assertFalse(api_version_request.is_supported(
            req, min_version='2.6'))

    def test_is_supported_max_version(self):
        req = fakes.HTTPRequest.blank(self.base_path, version='2.5')

        self.assertFalse(api_version_request.is_supported(
            req, max_version='2.4'))
        self.assertTrue(api_version_request.is_supported(
            req, max_version='2.5'))
        self.assertTrue(api_version_request.is_supported(
            req, max_version='2.6'))

    def test_is_supported_min_and_max_version(self):
        req = fakes.HTTPRequest.blank(self.base_path, version='2.5')

        self.assertFalse(api_version_request.is_supported(
            req, min_version='2.3', max_version='2.4'))
        self.assertTrue(api_version_request.is_supported(
            req, min_version='2.3', max_version='2.5'))
        self.assertTrue(api_version_request.is_supported(
            req, min_version='2.3', max_version='2.7'))
        self.assertTrue(api_version_request.is_supported(
            req, min_version='2.5', max_version='2.7'))
        self.assertFalse(api_version_request.is_supported(
            req, min_version='2.6', max_version='2.7'))
        self.assertTrue(api_version_request.is_supported(
            req, min_version='2.5', max_version='2.5'))
        self.assertFalse(api_version_request.is_supported(
            req, min_version='2.10', max_version='2.1'))



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from nova.policies import base


POLICY_ROOT = 'os_compute_api:os-quota-sets:%s'


quota_sets_policies = [
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'update',
        check_str=base.PROJECT_ADMIN,
        description="Update the quotas",
        operations=[
            {
                'method': 'PUT',
                'path': '/os-quota-sets/{tenant_id}'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'defaults',
        check_str=base.RULE_ANY,
        description="List default quotas",
        operations=[
            {
                'method': 'GET',
                'path': '/os-quota-sets/{tenant_id}/defaults'
            }
        ],
        scope_types=['system', 'project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'show',
        # TODO(gmann): Until we have domain admin or so to get other project's
        # data, allow admin role(with scope check it will be project admin) to
        # get other project quota. We cannot use PROJECT_ADMIN here as
        # project_id passed in request url is used as policy targets which
        # would not match with context's project_id fetched for rule
        # PROJECT_ADMIN check.
        check_str='(' + base.PROJECT_READER + ') or role:admin',
        description="Show a quota",
        operations=[
            {
                'method': 'GET',
                'path': '/os-quota-sets/{tenant_id}'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'delete',
        check_str=base.PROJECT_ADMIN,
        description="Revert quotas to defaults",
        operations=[
            {
                'method': 'DELETE',
                'path': '/os-quota-sets/{tenant_id}'
            }
        ],
        scope_types=['project']),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'detail',
        # TODO(gmann): Until we have domain admin or so to get other project's
        # data, allow admin role(with scope check it will be project admin) to
        # get other project quota.
        check_str='(' + base.PROJECT_READER + ') or role:admin',
        description="Show the detail of quota",
        operations=[
            {
                'method': 'GET',
                'path': '/os-quota-sets/{tenant_id}/detail'
            }
        ],
        scope_types=['project']),
]


def list_rules():
    return quota_sets_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

import fixtures
from oslo_config import cfg
from oslo_log import log as logging

from nova import exception


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def _get_device_profile(dp_name, trait):
    dp = {
        'fakedev-dp': [
            {
                'name': 'fakedev-dp',
                'uuid': 'cbec22f3-ac29-444e-b4bb-98509f32faae',
                'groups': [
                    {
                        'resources:FPGA': 1,
                        'trait:' + trait: 'required',
                    },
                ],
                # Skipping links key in Cyborg API return value
            },
            ],
        'fakedev-dp-port': [
            {
                'name': 'fakedev-dp',
                'uuid': 'cbec22f3-ac29-444e-b4bb-98509f32faae',
                'groups': [
                    {
                        'resources:FPGA': 1,
                        'trait:' + trait: 'required',
                    },
                ],
                # Skipping links key in Cyborg API return value
            },
         ],
         'fakedev-dp-multi': [
            {
                'name': 'fakedev-dp-multi',
                'uuid': 'cbec22f3-ac29-444e-b4bb-98509f32faae',
                'groups': [
                    {
                        'resources:FPGA': 2,
                        'resources:FPGA2': 1,
                        'trait:' + trait: 'required',
                    },
                ],
                # Skipping links key in Cyborg API return value
            },
          ],
        }
    return dp[dp_name]


def get_arqs(dp_name):
    # prepare fixture arqs and bound info
    arqs = [
        {
        'uuid': 'b59d34d3-787b-4fb0-a6b9-019cd81172f8',
        'device_profile_name': dp_name,
        'device_profile_group_id': 0,
        'state': 'Initial',
        'device_rp_uuid': None,
        'hostname': None,
        'instance_uuid': None,
        'attach_handle_info': {},
        'attach_handle_type': '',
        },
        {'uuid': '73d5f9f3-23e9-4b45-909a-e8a1db4cf24c',
        'device_profile_name': dp_name,
        'device_profile_group_id': 0,
        'state': 'Initial',
        'device_rp_uuid': None,
        'hostname': None,
        'instance_uuid': None,
        'attach_handle_info': {},
        'attach_handle_type': '',
        },
        {'uuid': '69b83caf-dd1c-493d-8796-40af5a16e3f6',
        'device_profile_name': dp_name,
        'device_profile_group_id': 0,
        'state': 'Initial',
        'device_rp_uuid': None,
        'hostname': None,
        'instance_uuid': None,
        'attach_handle_info': {},
        'attach_handle_type': '',
        },
        {'uuid': 'e5fc1da7-216b-4102-a50d-43ba77bcacf7',
        'device_profile_name': dp_name,
        'device_profile_group_id': 0,
        'state': 'Initial',
        'device_rp_uuid': None,
        'hostname': None,
        'instance_uuid': None,
        'attach_handle_info': {},
        'attach_handle_type': '',
        }
    ]
    # arqs bound info
    attach_handle_list = [
        {
            'bus': '0c',
            'device': '0',
            'domain': '0000',
            'function': '1',
            'physical_network': 'PHYNET1'
        },
        {
            'bus': '0c',
            'device': '0',
            'domain': '0000',
            'function': '2',
            'physical_network': 'PHYNET1'
        },
        {
            'bus': '0c',
            'device': '0',
            'domain': '0000',
            'function': '3',
            'physical_network': 'PHYNET1'
        },
        {
            'bus': '0c',
            'device': '0',
            'domain': '0000',
            'function': '4',
            'physical_network': 'PHYNET1'
        }
    ]

    bound_arqs = []
    # combine bond info to arq generating a bonded arqs list
    for idx, arq in enumerate(arqs):
        bound_arq = copy.deepcopy(arq)
        bound_arq.update(
            {'state': 'Bound',
            'attach_handle_type': 'TEST_PCI',
            'attach_handle_info': attach_handle_list[idx]},
            )
        bound_arqs.append(bound_arq)
    return arqs, bound_arqs


class CyborgFixture(fixtures.Fixture):
    """Fixture that mocks Cyborg APIs used by nova/accelerator/cyborg.py"""

    dp_name = 'fakedev-dp'
    trait = 'CUSTOM_FAKE_DEVICE'
    arq_list, bound_arq_list = copy.deepcopy(get_arqs(dp_name))
    arq_uuids = []
    for arq in arq_list:
        arq_uuids.append(arq["uuid"])
    call_create_arq_count = 0

    # NOTE(Sundar): The bindings passed to the fake_bind_arqs() from the
    # conductor are indexed by ARQ UUID and include the host name, device
    # RP UUID and instance UUID. (See params to fake_bind_arqs below.)
    #
    # Later, when the compute manager calls fake_get_arqs_for_instance() with
    # the instance UUID, the returned ARQs must contain the host name and
    # device RP UUID. But these can vary from test to test.
    #
    # So, fake_bind_arqs() below takes bindings indexed by ARQ UUID and
    # converts them to bindings indexed by instance UUID, which are then
    # stored in the dict below. This dict looks like:
    # { $instance_uuid: [
    #      {'hostname': $hostname,
    #       'device_rp_uuid': $device_rp_uuid,
    #       'arq_uuid': $arq_uuid
    #      }
    #   ]
    # }
    # Since it is indexed by instance UUID, and that is presumably unique
    # across concurrently executing tests, this should be safe for
    # concurrent access.

    def setUp(self):
        super().setUp()
        self.mock_get_dp = self.useFixture(fixtures.MockPatch(
            'nova.accelerator.cyborg._CyborgClient._get_device_profile_list',
            side_effect=self.fake_get_device_profile_list)).mock
        self.mock_create_arqs = self.useFixture(fixtures.MockPatch(
            'nova.accelerator.cyborg._CyborgClient.create_arqs',
            side_effect=self.fake_create_arqs)).mock
        self.mock_bind_arqs = self.useFixture(fixtures.MockPatch(
            'nova.accelerator.cyborg._CyborgClient.bind_arqs',
            side_effect=self.fake_bind_arqs)).mock
        self.mock_get_arqs = self.useFixture(fixtures.MockPatch(
            'nova.accelerator.cyborg._CyborgClient.'
            'get_arqs_for_instance',
            side_effect=self.fake_get_arqs_for_instance)).mock
        self.mock_del_arqs = self.useFixture(fixtures.MockPatch(
            'nova.accelerator.cyborg._CyborgClient.'
            'delete_arqs_for_instance',
            side_effect=self.fake_delete_arqs_for_instance)).mock
        self.mock_get_arq_by_uuid = self.useFixture(fixtures.MockPatch(
            'nova.accelerator.cyborg._CyborgClient.'
            'get_arq_by_uuid',
            side_effect=self.fake_get_arq_by_uuid)).mock
        self.mock_get_arq_device_rp_uuid = self.useFixture(fixtures.MockPatch(
            'nova.accelerator.cyborg._CyborgClient.'
            'get_arq_device_rp_uuid',
            side_effect=self.fake_get_arq_device_rp_uuid)).mock
        self.mock_create_arq_and_match_rp = self.useFixture(fixtures.MockPatch(
            'nova.accelerator.cyborg._CyborgClient.'
            'create_arqs_and_match_resource_providers',
            side_effect=self.fake_create_arq_and_match_rp)).mock
        self.mock_fake_delete_arqs_by_uuid = self.useFixture(
            fixtures.MockPatch(
                'nova.accelerator.cyborg._CyborgClient.'
                'delete_arqs_by_uuid',
                side_effect=self.fake_delete_arqs_by_uuid)).mock

    def fake_get_device_profile_list(self, dp_name):
        return _get_device_profile(dp_name, self.trait)

    @staticmethod
    def fake_bind_arqs(bindings):
        """Simulate Cyborg ARQ bindings.

        Since Nova calls Cyborg for binding on per-instance basis, the
        instance UUIDs would be the same for all ARQs in a single call.

        This function converts bindings indexed by ARQ UUID to bindings
        indexed by instance UUID, so that fake_get_arqs_for_instance can
        retrieve them later.

        :param bindings:
               { "$arq_uuid": {
                     "hostname": STRING
                     "device_rp_uuid": UUID
                     "instance_uuid": UUID
                  },
                  ...
                }
        :returns: None
        """
        if bindings.keys() and CyborgFixture.arq_uuids is None:
            LOG.error("ARQ not found")
            raise exception.AcceleratorRequestOpFailed()
        for arq_uuid, binding in bindings.items():
            for bound_arq in CyborgFixture.bound_arq_list:
                if arq_uuid == bound_arq["uuid"]:
                    bound_arq["hostname"] = binding["hostname"]
                    bound_arq["instance_uuid"] = binding["instance_uuid"]
                    bound_arq["device_rp_uuid"] = binding["device_rp_uuid"]
                    break

    @staticmethod
    def fake_get_arqs_for_instance(instance_uuid, only_resolved=False):
        """Get list of bound ARQs for this instance.

           This function uses bindings indexed by instance UUID to
           populate the bound ARQ templates in CyborgFixture.bound_arq_list.
        """
        bound_arq_list = copy.deepcopy(CyborgFixture.bound_arq_list)
        instance_bound_arqs = []
        for arq in bound_arq_list:
            if arq["instance_uuid"] == instance_uuid:
                instance_bound_arqs.append(arq)
        return instance_bound_arqs

    def fake_get_arq_by_uuid(self, uuid):
        for arq in self.arq_list:
            if uuid == arq['uuid']:
                return arq
        return None

    def fake_delete_arqs_for_instance(self, instance_uuid):
        # clean up arq binding info while delete arqs
        for arq in self.bound_arq_list:
            if arq["instance_uuid"] == instance_uuid:
                arq["instance_uuid"] = None
                arq["hostname"] = None
                arq["device_rp_uuid"] = None

    def fake_create_arq_and_match_rp(self,
        dp_name, rg_rp_map=None, owner=None):
        # sync the device_rp_uuid to fake arq
        arqs = self.fake_create_arqs(dp_name)
        for arq in arqs:
            dp_group_id = arq['device_profile_group_id']
            requester_id = ("device_profile_" + str(dp_group_id) +
                (str(owner) if owner else ""))
            arq["device_rp_uuid"] = rg_rp_map[requester_id][0]
        return arqs

    def fake_create_arqs(self, dp_name):
        index = self.call_create_arq_count
        self.call_create_arq_count += 1
        if index < len(self.arq_list):
            return [self.arq_list[index]]
        else:
            return None

    def fake_get_arq_device_rp_uuid(self,
        arq_arg, rg_rp_map=None, port_id=None):
        # sync the device_rp_uuid to fake arq
        for arq in self.arq_list:
            if arq["uuid"] == arq_arg['uuid']:
                dp_group_id = arq['device_profile_group_id']
                requester_id = ("device_profile_" +
                    str(dp_group_id) + str(port_id))
                arq["device_rp_uuid"] = rg_rp_map[requester_id][0]
                return arq["device_rp_uuid"]
        return None

    def fake_delete_arqs_by_uuid(self, arq_uuids):
        # clean up arq binding info while delete arqs
        for arq_uuid in arq_uuids:
            for arq in self.bound_arq_list:
                if arq["uuid"] == arq_uuid:
                    arq["instance_uuid"] = None
                    arq["hostname"] = None
                    arq["device_rp_uuid"] = None



# Copyright 2020, Red Hat, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

from oslo_utils.fixture import uuidsentinel as uuids

from nova.tests.functional.api import client
from nova.tests.functional import integrated_helpers


class TestNonBootableImageMeta(integrated_helpers._IntegratedTestBase):
    """Regression test for bug 1895696

    This regression test asserts the behaviour of server creation requests when
    using an image with nonbootable properties either directly in the request
    or to create a volume that is then booted from.
    """

    microversion = 'latest'

    def setUp(self):
        super().setUp()

        # Add an image to the Glance fixture with cinder_encryption_key set
        timestamp = datetime.datetime(2011, 1, 1, 1, 2, 3)
        cinder_encrypted_image = {
            'id': uuids.cinder_encrypted_image_uuid,
            'name': 'cinder_encryption_key_image',
            'created_at': timestamp,
            'updated_at': timestamp,
            'deleted_at': None,
            'deleted': False,
            'status': 'active',
            'is_public': False,
            'container_format': 'ova',
            'disk_format': 'vhd',
            'size': '74185822',
            'min_ram': 0,
            'min_disk': 0,
            'protected': False,
            'visibility': 'public',
            'tags': [],
            'properties': {
                'cinder_encryption_key_id': uuids.cinder_encryption_key_id,
            }
        }
        self.glance.create(None, cinder_encrypted_image)

        # Mock out nova.volume.cinder.API.{create,get} so that when n-api
        # requests that c-api create a volume from the above image that the
        # response includes cinder_encryption_key_id in the
        # volume_image_metadata
        cinder_encrypted_volume = {
            'status': 'available',
            'display_name': 'cinder_encrypted_volume',
            'attach_status': 'detached',
            'id': uuids.cinder_encrypted_volume_uuid,
            'multiattach': False,
            'size': 1,
            'encrypted': True,
            'volume_image_metadata': {
                'cinder_encryption_key_id': uuids.cinder_encryption_key_id
            }
        }

        def fake_cinder_create(self_api, context, size, name, description,
                snapshot=None, image_id=None, volume_type=None, metadata=None,
                availability_zone=None):
            if image_id == uuids.cinder_encrypted_image_uuid:
                return cinder_encrypted_volume
        self.stub_out(
            'nova.volume.cinder.API.create', fake_cinder_create)

        def fake_cinder_get(self_api, context, volume_id, microversion=None):
            return cinder_encrypted_volume
        self.stub_out(
            'nova.volume.cinder.API.get', fake_cinder_get)

    def test_nonbootable_metadata_image_metadata(self):
        """Assert behaviour when booting from an encrypted image
        """
        server = self._build_server(
            name='test_nonbootable_metadata_bfv_image_metadata',
            image_uuid=uuids.cinder_encrypted_image_uuid,
            networks='none'
        )
        # NOTE(lyarwood): This should always fail as Nova will attempt to boot
        # directly from this encrypted image.
        ex = self.assertRaises(
            client.OpenStackApiException, self.api.post_server,
            {'server': server})
        self.assertEqual(400, ex.response.status_code)
        self.assertIn(
            "Direct booting of an image uploaded from an encrypted volume is "
            "unsupported", str(ex))

    def test_nonbootable_metadata_bfv_image_metadata(self):
        """Assert behaviour when n-api creates volume using an encrypted image
        """
        server = self._build_server(
            name='test_nonbootable_metadata_bfv_image_metadata',
            image_uuid='', networks='none'
        )
        # TODO(lyarwood): Merge this into _build_server
        server['block_device_mapping_v2'] = [{
            'source_type': 'image',
            'destination_type': 'volume',
            'boot_index': 0,
            'uuid': uuids.cinder_encrypted_image_uuid,
            'volume_size': 1,
        }]

        # Assert that this request is accepted and the server moves to ACTIVE
        server = self.api.post_server({'server': server})
        self._wait_for_state_change(server, 'ACTIVE')

    def test_nonbootable_metadata_bfv_volume_image_metadata(self):
        """Assert behaviour when c-api has created volume using encrypted image
        """
        server = self._build_server(
            name='test_nonbootable_metadata_bfv_volume_image_metadata',
            image_uuid='', networks='none'
        )
        # TODO(lyarwood): Merge this into _build_server
        server['block_device_mapping_v2'] = [{
            'source_type': 'volume',
            'destination_type': 'volume',
            'boot_index': 0,
            'uuid': uuids.cinder_encrypted_volume_uuid,
        }]

        # Assert that this request is accepted and the server moves to ACTIVE
        server = self.api.post_server({'server': server})
        self._wait_for_state_change(server, 'ACTIVE')



# Copyright 2015 Rackspace Hosting, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import fixtures

from nova import test
from nova.virt.disk.mount import block
from nova.virt.image import model as imgmodel


class LoopTestCase(test.NoDBTestCase):
    def setUp(self):
        super(LoopTestCase, self).setUp()

        device_path = '/dev/mapper/instances--instance-0000001_disk'
        self.image = imgmodel.LocalBlockImage(device_path)

    def test_get_dev(self):
        tempdir = self.useFixture(fixtures.TempDir()).path
        b = block.BlockMount(self.image, tempdir)

        self.assertTrue(b.get_dev())
        self.assertTrue(b.linked)
        self.assertEqual(self.image.path, b.device)

    def test_unget_dev(self):
        tempdir = self.useFixture(fixtures.TempDir()).path
        b = block.BlockMount(self.image, tempdir)

        b.unget_dev()

        self.assertIsNone(b.device)
        self.assertFalse(b.linked)



# Copyright (c) 2016 Intel, Inc.
# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

consoleauth_group = cfg.OptGroup(
    name='consoleauth',
    title='Console auth options')

consoleauth_opts = [
    cfg.IntOpt('token_ttl',
        default=600,
        min=0,
        deprecated_name='console_token_ttl',
        deprecated_group='DEFAULT',
        help="""
The lifetime of a console auth token (in seconds).

A console auth token is used in authorizing console access for a user.
Once the auth token time to live count has elapsed, the token is
considered expired.  Expired tokens are then deleted.
""")
]


def register_opts(conf):
    conf.register_group(consoleauth_group)
    conf.register_opts(consoleauth_opts, group=consoleauth_group)


def list_opts():
    return {consoleauth_group: consoleauth_opts}



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy
import mock
from oslo_limit import fixture as limit_fixture
import webob

from nova.api.openstack.compute import quota_classes \
       as quota_classes_v21
from nova import exception
from nova.limit import local as local_limit
from nova.limit import placement as placement_limit
from nova import objects
from nova import test
from nova.tests.unit.api.openstack import fakes


class QuotaClassSetsTestV21(test.TestCase):
    validation_error = exception.ValidationError
    api_version = '2.1'
    quota_resources = {'metadata_items': 128,
                       'ram': 51200, 'floating_ips': -1,
                       'fixed_ips': -1, 'instances': 10,
                       'injected_files': 5, 'cores': 20,
                       'injected_file_content_bytes': 10240,
                       'security_groups': -1,
                       'security_group_rules': -1, 'key_pairs': 100,
                       'injected_file_path_bytes': 255}
    filtered_quotas = None

    def quota_set(self, class_name):
        quotas = copy.deepcopy(self.quota_resources)
        quotas['id'] = class_name
        return {'quota_class_set': quotas}

    def setUp(self):
        super(QuotaClassSetsTestV21, self).setUp()
        self.req = fakes.HTTPRequest.blank('', version=self.api_version)
        self._setup()

    def _setup(self):
        self.controller = quota_classes_v21.QuotaClassSetsController()

    def _check_filtered_extended_quota(self, quota_set):
        self.assertNotIn('server_groups', quota_set)
        self.assertNotIn('server_group_members', quota_set)
        self.assertEqual(-1, quota_set['floating_ips'])
        self.assertEqual(-1, quota_set['fixed_ips'])
        self.assertEqual(-1, quota_set['security_groups'])
        self.assertEqual(-1, quota_set['security_group_rules'])

    def test_format_quota_set(self):
        quota_set = self.controller._format_quota_set('test_class',
                                                      self.quota_resources,
                                                      self.filtered_quotas)
        qs = quota_set['quota_class_set']

        self.assertEqual(qs['id'], 'test_class')
        for resource, value in self.quota_resources.items():
            self.assertEqual(value, qs[resource])
        if self.filtered_quotas:
            for resource in self.filtered_quotas:
                self.assertNotIn(resource, qs)
        self._check_filtered_extended_quota(qs)

    def test_quotas_show(self):
        res_dict = self.controller.show(self.req, 'test_class')

        self.assertEqual(res_dict, self.quota_set('test_class'))

    def test_quotas_update(self):
        expected_body = {'quota_class_set': self.quota_resources}
        request_quota_resources = copy.deepcopy(self.quota_resources)
        request_quota_resources['server_groups'] = 10
        request_quota_resources['server_group_members'] = 10
        request_body = {'quota_class_set': request_quota_resources}
        res_dict = self.controller.update(self.req, 'test_class',
                                          body=request_body)

        self.assertEqual(res_dict, expected_body)

    def test_quotas_update_with_empty_body(self):
        body = {}
        self.assertRaises(self.validation_error, self.controller.update,
                          self.req, 'test_class', body=body)

    def test_quotas_update_with_invalid_integer(self):
        body = {'quota_class_set': {'instances': 2 ** 31 + 1}}
        self.assertRaises(self.validation_error, self.controller.update,
                          self.req, 'test_class', body=body)

    def test_quotas_update_with_long_quota_class_name(self):
        name = 'a' * 256
        body = {'quota_class_set': {'instances': 10}}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, name, body=body)

    def test_quotas_update_with_non_integer(self):
        body = {'quota_class_set': {'instances': "abc"}}
        self.assertRaises(self.validation_error, self.controller.update,
                          self.req, 'test_class', body=body)

        body = {'quota_class_set': {'instances': 50.5}}
        self.assertRaises(self.validation_error, self.controller.update,
                          self.req, 'test_class', body=body)

        body = {'quota_class_set': {
                'instances': u'\u30aa\u30fc\u30d7\u30f3'}}
        self.assertRaises(self.validation_error, self.controller.update,
                          self.req, 'test_class', body=body)

    def test_quotas_update_with_unsupported_quota_class(self):
        body = {'quota_class_set': {'instances': 50, 'cores': 50,
                                    'ram': 51200, 'unsupported': 12}}
        self.assertRaises(self.validation_error, self.controller.update,
                          self.req, 'test_class', body=body)


class QuotaClassSetsTestV250(QuotaClassSetsTestV21):
    api_version = '2.50'
    quota_resources = {'metadata_items': 128,
                       'ram': 51200, 'instances': 10,
                       'injected_files': 5, 'cores': 20,
                       'injected_file_content_bytes': 10240,
                       'key_pairs': 100,
                       'injected_file_path_bytes': 255,
                       'server_groups': 10,
                       'server_group_members': 10}
    filtered_quotas = quota_classes_v21.FILTERED_QUOTAS_2_50

    def _check_filtered_extended_quota(self, quota_set):
        self.assertEqual(10, quota_set['server_groups'])
        self.assertEqual(10, quota_set['server_group_members'])
        for resource in self.filtered_quotas:
            self.assertNotIn(resource, quota_set)

    def test_quotas_update_with_filtered_quota(self):
        for resource in self.filtered_quotas:
            body = {'quota_class_set': {resource: 10}}
            self.assertRaises(self.validation_error, self.controller.update,
                              self.req, 'test_class', body=body)


class QuotaClassSetsTestV257(QuotaClassSetsTestV250):
    api_version = '2.57'

    def setUp(self):
        super(QuotaClassSetsTestV257, self).setUp()
        for resource in quota_classes_v21.FILTERED_QUOTAS_2_57:
            self.quota_resources.pop(resource, None)
        self.filtered_quotas.extend(quota_classes_v21.FILTERED_QUOTAS_2_57)


class NoopQuotaClassesTest(test.NoDBTestCase):
    quota_driver = "nova.quota.NoopQuotaDriver"

    def setUp(self):
        super(NoopQuotaClassesTest, self).setUp()
        self.flags(driver=self.quota_driver, group="quota")
        self.controller = quota_classes_v21.QuotaClassSetsController()

    def test_show_v21(self):
        req = fakes.HTTPRequest.blank("")
        response = self.controller.show(req, "test_class")
        expected_response = {
            'quota_class_set': {
                'id': 'test_class',
                'cores': -1,
                'fixed_ips': -1,
                'floating_ips': -1,
                'injected_file_content_bytes': -1,
                'injected_file_path_bytes': -1,
                'injected_files': -1,
                'instances': -1,
                'key_pairs': -1,
                'metadata_items': -1,
                'ram': -1,
                'security_group_rules': -1,
                'security_groups': -1
            }
        }
        self.assertEqual(expected_response, response)

    def test_show_v257(self):
        req = fakes.HTTPRequest.blank("", version='2.57')
        response = self.controller.show(req, "default")
        expected_response = {
            'quota_class_set': {
                'id': 'default',
                'cores': -1,
                'instances': -1,
                'key_pairs': -1,
                'metadata_items': -1,
                'ram': -1,
                'server_group_members': -1,
                'server_groups': -1,
            }
        }
        self.assertEqual(expected_response, response)

    def test_update_v21_still_rejects_badrequests(self):
        req = fakes.HTTPRequest.blank("")
        body = {'quota_class_set': {'instances': 50, 'cores': 50,
                                    'ram': 51200, 'unsupported': 12}}
        self.assertRaises(exception.ValidationError, self.controller.update,
                          req, 'test_class', body=body)

    @mock.patch.object(objects.Quotas, "update_class")
    def test_update_v21(self, mock_update):
        req = fakes.HTTPRequest.blank("")
        body = {'quota_class_set': {'ram': 51200}}
        response = self.controller.update(req, 'default', body=body)
        expected_response = {
            'quota_class_set': {
                'cores': -1,
                'fixed_ips': -1,
                'floating_ips': -1,
                'injected_file_content_bytes': -1,
                'injected_file_path_bytes': -1,
                'injected_files': -1,
                'instances': -1,
                'key_pairs': -1,
                'metadata_items': -1,
                'ram': -1,
                'security_group_rules': -1,
                'security_groups': -1
            }
        }
        self.assertEqual(expected_response, response)
        mock_update.assert_called_once_with(req.environ['nova.context'],
                                            "default", "ram", 51200)

    @mock.patch.object(objects.Quotas, "update_class")
    def test_update_v257(self, mock_update):
        req = fakes.HTTPRequest.blank("", version='2.57')
        body = {'quota_class_set': {'ram': 51200}}
        response = self.controller.update(req, 'default', body=body)
        expected_response = {
            'quota_class_set': {
                'cores': -1,
                'instances': -1,
                'key_pairs': -1,
                'metadata_items': -1,
                'ram': -1,
                'server_group_members': -1,
                'server_groups': -1,
            }
        }
        self.assertEqual(expected_response, response)
        mock_update.assert_called_once_with(req.environ['nova.context'],
                                            "default", "ram", 51200)


class UnifiedLimitsQuotaClassesTest(NoopQuotaClassesTest):
    quota_driver = "nova.quota.UnifiedLimitsDriver"

    def setUp(self):
        super(UnifiedLimitsQuotaClassesTest, self).setUp()
        # Set server_groups so all config options get a different value
        # but we also test as much as possible with the default config
        self.flags(driver="nova.quota.UnifiedLimitsDriver", group='quota')
        reglimits = {local_limit.SERVER_METADATA_ITEMS: 128,
                     local_limit.INJECTED_FILES: 5,
                     local_limit.INJECTED_FILES_CONTENT: 10 * 1024,
                     local_limit.INJECTED_FILES_PATH: 255,
                     local_limit.KEY_PAIRS: 100,
                     local_limit.SERVER_GROUPS: 12,
                     local_limit.SERVER_GROUP_MEMBERS: 10}
        self.useFixture(limit_fixture.LimitFixture(reglimits, {}))

    @mock.patch.object(placement_limit, "get_legacy_default_limits")
    def test_show_v21(self, mock_default):
        mock_default.return_value = {"instances": 1, "cores": 2, "ram": 3}
        req = fakes.HTTPRequest.blank("")
        response = self.controller.show(req, "test_class")
        expected_response = {
            'quota_class_set': {
                'id': 'test_class',
                'cores': 2,
                'fixed_ips': -1,
                'floating_ips': -1,
                'ram': 3,
                'injected_file_content_bytes': 10240,
                'injected_file_path_bytes': 255,
                'injected_files': 5,
                'instances': 1,
                'key_pairs': 100,
                'metadata_items': 128,
                'security_group_rules': -1,
                'security_groups': -1,
            }
        }
        self.assertEqual(expected_response, response)

    @mock.patch.object(placement_limit, "get_legacy_default_limits")
    def test_show_v257(self, mock_default):
        mock_default.return_value = {"instances": 1, "cores": 2, "ram": 3}
        req = fakes.HTTPRequest.blank("", version='2.57')
        response = self.controller.show(req, "default")
        expected_response = {
            'quota_class_set': {
                'id': 'default',
                'cores': 2,
                'instances': 1,
                'ram': 3,
                'key_pairs': 100,
                'metadata_items': 128,
                'server_group_members': 10,
                'server_groups': 12,
            }
        }
        self.assertEqual(expected_response, response)

    def test_update_still_rejects_badrequests(self):
        req = fakes.HTTPRequest.blank("")
        body = {'quota_class_set': {'instances': 50, 'cores': 50,
                                    'ram': 51200, 'unsupported': 12}}
        self.assertRaises(exception.ValidationError, self.controller.update,
                          req, 'test_class', body=body)

    @mock.patch.object(placement_limit, "get_legacy_default_limits")
    @mock.patch.object(objects.Quotas, "update_class")
    def test_update_v21(self, mock_update, mock_default):
        mock_default.return_value = {"instances": 1, "cores": 2, "ram": 3}
        req = fakes.HTTPRequest.blank("")
        body = {'quota_class_set': {'ram': 51200}}
        response = self.controller.update(req, 'default', body=body)
        expected_response = {
            'quota_class_set': {
                'cores': 2,
                'fixed_ips': -1,
                'floating_ips': -1,
                'injected_file_content_bytes': 10240,
                'injected_file_path_bytes': 255,
                'injected_files': 5,
                'instances': 1,
                'key_pairs': 100,
                'metadata_items': 128,
                'ram': 3,
                'security_group_rules': -1,
                'security_groups': -1
            }
        }
        self.assertEqual(expected_response, response)
        # TODO(johngarbutt) we should be proxying to keystone
        self.assertEqual(0, mock_update.call_count)

    @mock.patch.object(placement_limit, "get_legacy_default_limits")
    @mock.patch.object(objects.Quotas, "update_class")
    def test_update_v257(self, mock_update, mock_default):
        mock_default.return_value = {"instances": 1, "cores": 2, "ram": 3}
        req = fakes.HTTPRequest.blank("", version='2.57')
        body = {'quota_class_set': {'ram': 51200}}
        response = self.controller.update(req, 'default', body=body)
        expected_response = {
            'quota_class_set': {
                'cores': 2,
                'instances': 1,
                'ram': 3,
                'key_pairs': 100,
                'metadata_items': 128,
                'server_group_members': 10,
                'server_groups': 12,
            }
        }
        self.assertEqual(expected_response, response)
        # TODO(johngarbutt) we should be proxying to keystone
        self.assertEqual(0, mock_update.call_count)



# Copyright 2017 Huawei Technologies Co.,LTD.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from nova.api.validation import parameter_types

list_query_schema_v20 = {
    'type': 'object',
    'properties': {
        'hidden': parameter_types.common_query_param,
        'host': parameter_types.common_query_param,
        'instance_uuid': parameter_types.common_query_param,
        'source_compute': parameter_types.common_query_param,
        'status': parameter_types.common_query_param,
        'migration_type': parameter_types.common_query_param,
    },
    # For backward compatible changes
    'additionalProperties': True
}

list_query_params_v259 = copy.deepcopy(list_query_schema_v20)
list_query_params_v259['properties'].update({
    # The 2.59 microversion added support for paging by limit and marker
    # and filtering by changes-since.
    'limit': parameter_types.single_param(
        parameter_types.non_negative_integer),
    'marker': parameter_types.single_param({'type': 'string'}),
    'changes-since': parameter_types.single_param(
        {'type': 'string', 'format': 'date-time'}),
})
list_query_params_v259['additionalProperties'] = False

list_query_params_v266 = copy.deepcopy(list_query_params_v259)
list_query_params_v266['properties'].update({
    'changes-before': parameter_types.single_param(
        {'type': 'string', 'format': 'date-time'}),
})

list_query_params_v280 = copy.deepcopy(list_query_params_v266)
list_query_params_v280['properties'].update({
    # The 2.80 microversion added support for filtering migrations
    # by user_id and/or project_id
    'user_id': parameter_types.single_param({'type': 'string'}),
    'project_id': parameter_types.single_param({'type': 'string'}),
})



#    Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import traceback

import mock
from oslo_utils import fixture as utils_fixture
from oslo_utils.fixture import uuidsentinel as uuids
from oslo_utils import timeutils

from nova.db.main import api as db
from nova import exception
from nova import objects
from nova.objects import instance_action
from nova import test
from nova.tests.unit.objects import test_objects


NOW = timeutils.utcnow().replace(microsecond=0)
fake_action = {
    'created_at': NOW,
    'deleted_at': None,
    'updated_at': None,
    'deleted': False,
    'id': 123,
    'action': 'fake-action',
    'instance_uuid': uuids.instance,
    'request_id': 'fake-request',
    'user_id': 'fake-user',
    'project_id': 'fake-project',
    'start_time': NOW,
    'finish_time': None,
    'message': 'foo',
}
fake_event = {
    'created_at': NOW,
    'deleted_at': None,
    'updated_at': None,
    'deleted': False,
    'id': 123,
    'event': 'fake-event',
    'action_id': 123,
    'start_time': NOW,
    'finish_time': None,
    'result': 'fake-result',
    'traceback': 'fake-tb',
    'host': 'fake-host',
    'details': None
}


class _TestInstanceActionObject(object):
    @mock.patch.object(db, 'action_get_by_request_id')
    def test_get_by_request_id(self, mock_get):
        context = self.context
        mock_get.return_value = fake_action
        action = instance_action.InstanceAction.get_by_request_id(
            context, 'fake-uuid', 'fake-request')
        self.compare_obj(action, fake_action)
        mock_get.assert_called_once_with(context,
            'fake-uuid', 'fake-request')

    def test_pack_action_start(self):
        values = instance_action.InstanceAction.pack_action_start(
            self.context, 'fake-uuid', 'fake-action')
        self.assertEqual(values['request_id'], self.context.request_id)
        self.assertEqual(values['user_id'], self.context.user_id)
        self.assertEqual(values['project_id'], self.context.project_id)
        self.assertEqual(values['instance_uuid'], 'fake-uuid')
        self.assertEqual(values['action'], 'fake-action')
        self.assertEqual(values['start_time'].replace(tzinfo=None),
                         self.context.timestamp)

    def test_pack_action_finish(self):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        values = instance_action.InstanceAction.pack_action_finish(
            self.context, 'fake-uuid')
        self.assertEqual(values['request_id'], self.context.request_id)
        self.assertEqual(values['instance_uuid'], 'fake-uuid')
        self.assertEqual(values['finish_time'].replace(tzinfo=None), NOW)

    @mock.patch.object(db, 'action_start')
    def test_action_start(self, mock_start):
        test_class = instance_action.InstanceAction
        expected_packed_values = test_class.pack_action_start(
            self.context, 'fake-uuid', 'fake-action')
        mock_start.return_value = fake_action
        action = instance_action.InstanceAction.action_start(
            self.context, 'fake-uuid', 'fake-action', want_result=True)
        mock_start.assert_called_once_with(self.context,
                                           expected_packed_values)
        self.compare_obj(action, fake_action)

    @mock.patch.object(db, 'action_start')
    def test_action_start_no_result(self, mock_start):
        test_class = instance_action.InstanceAction
        expected_packed_values = test_class.pack_action_start(
            self.context, 'fake-uuid', 'fake-action')
        mock_start.return_value = fake_action
        action = instance_action.InstanceAction.action_start(
            self.context, 'fake-uuid', 'fake-action', want_result=False)
        mock_start.assert_called_once_with(self.context,
                                           expected_packed_values)
        self.assertIsNone(action)

    @mock.patch.object(db, 'action_finish')
    def test_action_finish(self, mock_finish):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceAction
        expected_packed_values = test_class.pack_action_finish(
            self.context, 'fake-uuid')
        mock_finish.return_value = fake_action
        action = instance_action.InstanceAction.action_finish(
            self.context, 'fake-uuid', want_result=True)
        mock_finish.assert_called_once_with(self.context,
                                            expected_packed_values)
        self.compare_obj(action, fake_action)

    @mock.patch.object(db, 'action_finish')
    def test_action_finish_no_result(self, mock_finish):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceAction
        expected_packed_values = test_class.pack_action_finish(
            self.context, 'fake-uuid')
        mock_finish.return_value = fake_action
        action = instance_action.InstanceAction.action_finish(
            self.context, 'fake-uuid', want_result=False)
        mock_finish.assert_called_once_with(self.context,
                                            expected_packed_values)
        self.assertIsNone(action)

    @mock.patch.object(db, 'action_finish')
    @mock.patch.object(db, 'action_start')
    def test_finish(self, mock_start, mock_finish):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        expected_packed_action_start = {
            'request_id': self.context.request_id,
            'user_id': self.context.user_id,
            'project_id': self.context.project_id,
            'instance_uuid': uuids.instance,
            'action': 'fake-action',
            'start_time': self.context.timestamp,
            'updated_at': self.context.timestamp,
        }
        expected_packed_action_finish = {
            'request_id': self.context.request_id,
            'instance_uuid': uuids.instance,
            'finish_time': NOW,
            'updated_at': NOW,
        }
        mock_start.return_value = fake_action
        mock_finish.return_value = fake_action
        action = instance_action.InstanceAction.action_start(
            self.context, uuids.instance, 'fake-action')
        action.finish()
        mock_start.assert_called_once_with(self.context,
                                           expected_packed_action_start)
        mock_finish.assert_called_once_with(self.context,
                                           expected_packed_action_finish)
        self.compare_obj(action, fake_action)

    @mock.patch.object(db, 'actions_get')
    def test_get_list(self, mock_get):
        fake_actions = [dict(fake_action, id=1234),
                        dict(fake_action, id=5678)]
        mock_get.return_value = fake_actions
        obj_list = instance_action.InstanceActionList.get_by_instance_uuid(
            self.context, 'fake-uuid')
        for index, action in enumerate(obj_list):
            self.compare_obj(action, fake_actions[index])
        mock_get.assert_called_once_with(self.context, 'fake-uuid', None,
                                         None, None)

    def test_create_id_in_updates_error(self):
        action = instance_action.InstanceAction(self.context, id=1)
        ex = self.assertRaises(exception.ObjectActionError, action.create)
        self.assertIn('already created', str(ex))

    @mock.patch('nova.db.main.api.action_start')
    def test_create(self, mock_action_start):
        mock_action_start.return_value = fake_action
        action = instance_action.InstanceAction(self.context)
        expected_updates = action.obj_get_changes()
        action.create()
        mock_action_start.assert_called_once_with(
            self.context, expected_updates)
        self.compare_obj(action, fake_action)


class TestInstanceActionObject(test_objects._LocalTest,
                               _TestInstanceActionObject):
    pass


class TestRemoteInstanceActionObject(test_objects._RemoteTest,
                                     _TestInstanceActionObject):
    pass


class _TestInstanceActionEventObject(object):
    @mock.patch.object(db, 'action_event_get_by_id')
    def test_get_by_id(self, mock_get):
        mock_get.return_value = fake_event
        event = instance_action.InstanceActionEvent.get_by_id(
            self.context, 'fake-action-id', 'fake-event-id')
        self.compare_obj(event, fake_event)
        mock_get.assert_called_once_with(self.context,
            'fake-action-id', 'fake-event-id')

    @mock.patch.object(db, 'action_event_start')
    def test_event_start(self, mock_start):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceActionEvent
        expected_packed_values = test_class.pack_action_event_start(
            self.context, 'fake-uuid', 'fake-event')
        mock_start.return_value = fake_event
        event = instance_action.InstanceActionEvent.event_start(
            self.context, 'fake-uuid', 'fake-event', want_result=True)
        mock_start.assert_called_once_with(self.context,
                                           expected_packed_values)
        self.compare_obj(event, fake_event)

    @mock.patch.object(db, 'action_event_start')
    def test_event_start_no_result(self, mock_start):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceActionEvent
        expected_packed_values = test_class.pack_action_event_start(
            self.context, 'fake-uuid', 'fake-event')
        mock_start.return_value = fake_event
        event = instance_action.InstanceActionEvent.event_start(
            self.context, 'fake-uuid', 'fake-event', want_result=False)
        mock_start.assert_called_once_with(self.context,
                                           expected_packed_values)
        self.assertIsNone(event)

    @mock.patch.object(db, 'action_event_finish')
    def test_event_finish(self, mock_finish):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceActionEvent
        expected_packed_values = test_class.pack_action_event_finish(
            self.context, 'fake-uuid', 'fake-event')
        expected_packed_values['finish_time'] = NOW
        self.assertNotIn('details', expected_packed_values)
        mock_finish.return_value = fake_event
        event = instance_action.InstanceActionEvent.event_finish(
            self.context, 'fake-uuid', 'fake-event', want_result=True)
        mock_finish.assert_called_once_with(self.context,
                                            expected_packed_values)
        self.compare_obj(event, fake_event)

    @mock.patch.object(db, 'action_event_finish')
    def test_event_finish_no_result(self, mock_finish):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceActionEvent
        expected_packed_values = test_class.pack_action_event_finish(
            self.context, 'fake-uuid', 'fake-event')
        expected_packed_values['finish_time'] = NOW
        mock_finish.return_value = fake_event
        event = instance_action.InstanceActionEvent.event_finish(
            self.context, 'fake-uuid', 'fake-event', want_result=False)
        mock_finish.assert_called_once_with(self.context,
                                            expected_packed_values)
        self.assertIsNone(event)

    @mock.patch.object(traceback, 'format_tb')
    @mock.patch.object(db, 'action_event_finish')
    def test_event_finish_with_failure(self, mock_finish, mock_tb):
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceActionEvent
        # The NovaException message will get formatted for the 'details' field.
        exc_val = exception.NoValidHost(reason='some error')
        expected_packed_values = test_class.pack_action_event_finish(
            self.context, 'fake-uuid', 'fake-event', exc_val, 'fake-tb')
        expected_packed_values['finish_time'] = NOW
        self.assertEqual(exc_val.format_message(),
                         expected_packed_values['details'])

        fake_event_with_details = copy.deepcopy(fake_event)
        fake_event_with_details['details'] = expected_packed_values['details']
        mock_finish.return_value = fake_event_with_details
        event = test_class.event_finish_with_failure(
            self.context, 'fake-uuid', 'fake-event', exc_val, 'fake-tb',
            want_result=True)
        mock_finish.assert_called_once_with(self.context,
                                            expected_packed_values)
        self.compare_obj(event, fake_event_with_details)
        mock_tb.assert_not_called()

    @mock.patch.object(traceback, 'format_tb')
    @mock.patch.object(db, 'action_event_finish')
    def test_event_finish_with_failure_legacy(self, mock_finish, mock_tb):
        # Tests that exc_tb is serialized when it's not a string type.
        mock_tb.return_value = 'fake-tb'
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceActionEvent
        # A non-NovaException will use the exception class name for
        # the 'details' field.
        exc_val = test.TestingException('non-nova-error')
        expected_packed_values = test_class.pack_action_event_finish(
            self.context, 'fake-uuid', 'fake-event', exc_val, 'fake-tb')
        expected_packed_values['finish_time'] = NOW
        self.assertEqual('TestingException', expected_packed_values['details'])

        fake_event_with_details = copy.deepcopy(fake_event)
        fake_event_with_details['details'] = expected_packed_values['details']
        mock_finish.return_value = fake_event_with_details
        fake_tb = mock.sentinel.fake_tb
        event = test_class.event_finish_with_failure(
            self.context, 'fake-uuid', 'fake-event', exc_val=exc_val,
            exc_tb=fake_tb, want_result=True)
        # When calling event_finish_with_failure and using exc_val as a kwarg
        # serialize_args will convert exc_val to non-nova exception class name
        # form before it reaches event_finish_with_failure.
        mock_finish.assert_called_once_with(self.context,
                                            expected_packed_values)
        self.compare_obj(event, fake_event_with_details)
        mock_tb.assert_called_once_with(fake_tb)

    @mock.patch.object(db, 'action_event_finish')
    def test_event_finish_with_failure_legacy_unicode(self, mock_finish):
        # Tests that traceback.format_tb is not called when exc_tb is unicode.
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceActionEvent
        expected_packed_values = test_class.pack_action_event_finish(
            self.context, 'fake-uuid', 'fake-event', 'val', 'fake-tb')
        expected_packed_values['finish_time'] = NOW

        mock_finish.return_value = fake_event
        event = test_class.event_finish_with_failure(
            self.context, 'fake-uuid', 'fake-event', exc_val='val',
            exc_tb='fake-tb', want_result=True)
        mock_finish.assert_called_once_with(self.context,
                                            expected_packed_values)
        self.compare_obj(event, fake_event)

    @mock.patch.object(traceback, 'format_tb')
    @mock.patch.object(db, 'action_event_finish')
    def test_event_finish_with_failure_no_result(self, mock_finish, mock_tb):
        # Tests that traceback.format_tb is not called when exc_tb is a str
        # and want_result is False, so no event should come back.
        mock_tb.return_value = 'fake-tb'
        self.useFixture(utils_fixture.TimeFixture(NOW))
        test_class = instance_action.InstanceActionEvent
        expected_packed_values = test_class.pack_action_event_finish(
            self.context, 'fake-uuid', 'fake-event', 'val', 'fake-tb')
        expected_packed_values['finish_time'] = NOW

        mock_finish.return_value = fake_event
        event = test_class.event_finish_with_failure(
            self.context, 'fake-uuid', 'fake-event', 'val', 'fake-tb',
            want_result=False)
        mock_finish.assert_called_once_with(self.context,
                                            expected_packed_values)
        self.assertIsNone(event)
        self.assertFalse(mock_tb.called)

    @mock.patch.object(db, 'action_events_get')
    def test_get_by_action(self, mock_get):
        fake_events = [dict(fake_event, id=1234),
                       dict(fake_event, id=5678)]
        mock_get.return_value = fake_events
        obj_list = instance_action.InstanceActionEventList.get_by_action(
            self.context, 'fake-action-id')
        for index, event in enumerate(obj_list):
            self.compare_obj(event, fake_events[index])
        mock_get.assert_called_once_with(self.context, 'fake-action-id')

    @mock.patch('nova.objects.instance_action.InstanceActionEvent.'
                'pack_action_event_finish')
    @mock.patch('traceback.format_tb')
    def test_event_finish_with_failure_serialized(self, mock_format,
                                                  mock_pack):
        mock_format.return_value = 'traceback'
        mock_pack.side_effect = test.TestingException
        exc = exception.NotFound()
        self.assertRaises(
            test.TestingException,
            instance_action.InstanceActionEvent.event_finish_with_failure,
            self.context, 'fake-uuid', 'fake-event',
            exc_val=exc,
            exc_tb=mock.sentinel.exc_tb)
        mock_pack.assert_called_once_with(self.context, 'fake-uuid',
                                          'fake-event',
                                          exc_val=exc.format_message(),
                                          exc_tb='traceback')
        mock_format.assert_called_once_with(mock.sentinel.exc_tb)

    def test_create_id_in_updates_error(self):
        event = instance_action.InstanceActionEvent(self.context, id=1)
        ex = self.assertRaises(
            exception.ObjectActionError, event.create,
            fake_action['instance_uuid'], fake_action['request_id'])
        self.assertIn('already created', str(ex))

    @mock.patch('nova.db.main.api.action_event_start')
    def test_create(self, mock_action_event_start):
        mock_action_event_start.return_value = fake_event
        event = instance_action.InstanceActionEvent(self.context)
        expected_updates = event.obj_get_changes()
        expected_updates['instance_uuid'] = fake_action['instance_uuid']
        expected_updates['request_id'] = fake_action['request_id']
        event.create(fake_action['instance_uuid'], fake_action['request_id'])
        mock_action_event_start.assert_called_once_with(
            self.context, expected_updates)
        self.compare_obj(event, fake_event)

    def test_obj_make_compatible(self):
        action_event_obj = objects.InstanceActionEvent(
            details=None,       # added in 1.4
            host='fake-host'    # added in 1.2
        )
        data = lambda x: x['nova_object.data']
        primitive = data(action_event_obj.obj_to_primitive(
            target_version='1.3'))
        self.assertIn('host', primitive)
        self.assertNotIn('details', primitive)


class TestInstanceActionEventObject(test_objects._LocalTest,
                                    _TestInstanceActionEventObject):
    pass


class TestRemoteInstanceActionEventObject(test_objects._RemoteTest,
                                          _TestInstanceActionEventObject):
    pass



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from nova.policies import base

POLICY_ROOT = 'os_compute_api:os-flavor-extra-specs:%s'

flavor_extra_specs_policies = [
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'show',
        check_str=base.PROJECT_READER_OR_ADMIN,
        description="Show an extra spec for a flavor",
        operations=[
            {
                'path': '/flavors/{flavor_id}/os-extra_specs/'
                        '{flavor_extra_spec_key}',
                'method': 'GET'
            }
        ],
        scope_types=['system', 'project']
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'create',
        check_str=base.ADMIN,
        description="Create extra specs for a flavor",
        operations=[
            {
                'path': '/flavors/{flavor_id}/os-extra_specs/',
                'method': 'POST'
            }
        ],
        scope_types=['system']
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'update',
        check_str=base.ADMIN,
        description="Update an extra spec for a flavor",
        operations=[
            {
                'path': '/flavors/{flavor_id}/os-extra_specs/'
                        '{flavor_extra_spec_key}',
                'method': 'PUT'
            }
        ],
        scope_types=['system']
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'delete',
        check_str=base.ADMIN,
        description="Delete an extra spec for a flavor",
        operations=[
            {
                'path': '/flavors/{flavor_id}/os-extra_specs/'
                        '{flavor_extra_spec_key}',
                'method': 'DELETE'
            }
        ],
        scope_types=['system']
    ),
    policy.DocumentedRuleDefault(
        name=POLICY_ROOT % 'index',
        check_str=base.PROJECT_READER_OR_ADMIN,
        description="List extra specs for a flavor. Starting with "
        "microversion 2.61, extra specs may be returned in responses "
        "for the flavor resource.",
        operations=[
            {
                'path': '/flavors/{flavor_id}/os-extra_specs/',
                'method': 'GET'
            },
            # Microversion 2.61 operations for flavors:
            {
                'path': '/flavors',
                'method': 'POST'
            },
            {
                'path': '/flavors/detail',
                'method': 'GET'
            },
            {
                'path': '/flavors/{flavor_id}',
                'method': 'GET'
            },
            {
                'path': '/flavors/{flavor_id}',
                'method': 'PUT'
            }
        ],
        scope_types=['system', 'project']
    ),
]


def list_rules():
    return flavor_extra_specs_policies



#
# Copyright 2013 - Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Test the base rpc API.
"""

from nova import baserpc
from nova.compute import rpcapi as compute_rpcapi
from nova import context
from nova import test


class BaseAPITestCase(test.TestCase):

    def setUp(self):
        super(BaseAPITestCase, self).setUp()
        self.user_id = 'fake'
        self.project_id = 'fake'
        self.context = context.RequestContext(self.user_id,
                                              self.project_id)
        self.compute = self.start_service('compute')
        self.base_rpcapi = baserpc.BaseAPI(compute_rpcapi.RPC_TOPIC)

    def test_ping(self):
        res = self.base_rpcapi.ping(self.context, 'foo')
        self.assertEqual({'service': 'compute', 'arg': 'foo'}, res)

    def test_get_backdoor_port(self):
        res = self.base_rpcapi.get_backdoor_port(self.context,
                self.compute.host)
        self.assertEqual(self.compute.backdoor_port, res)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from nova import context as nova_context
from nova import exception
from nova.objects import service
from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional import fixtures as func_fixtures
from nova.tests.functional import integrated_helpers


class ServiceTestCase(test.TestCase,
                      integrated_helpers.InstanceHelperMixin):
    """Contains scenarios for testing services.
    """

    def setUp(self):
        super(ServiceTestCase, self).setUp()
        # Use the standard fixtures.
        self.useFixture(nova_fixtures.RealPolicyFixture())
        self.useFixture(nova_fixtures.NeutronFixture(self))
        self.useFixture(nova_fixtures.GlanceFixture(self))
        self.useFixture(func_fixtures.PlacementFixture())

        # Start nova controller services.
        self.api = self.useFixture(nova_fixtures.OSAPIFixture(
            api_version='v2.1')).api
        self.start_service('conductor')
        self.scheduler = self.start_service('scheduler')
        # Our OSAPIFixture does not use a WSGIService, so just use the metadata
        # server fixture (which uses WSGIService) for testing.
        self.metadata = self.useFixture(
            nova_fixtures.OSMetadataServer()).metadata
        # Start one compute service.
        self.start_service('compute')

    def test_service_reset_resets_cell_cache(self):
        """Tests that the cell cache for database transaction context managers
        is cleared after a service reset (example scenario: SIGHUP).
        """
        server_req = self._build_server()
        server = self.api.post_server({'server': server_req})
        self._wait_for_state_change(server, 'ACTIVE')
        # Cell cache should be populated after creating a server.
        self.assertTrue(nova_context.CELL_CACHE)
        self.scheduler.reset()
        # Cell cache should be empty after the service reset.
        self.assertEqual({}, nova_context.CELL_CACHE)

        # Now test the WSGI service.
        server = self.api.post_server({'server': server_req})
        self._wait_for_state_change(server, 'ACTIVE')
        # Cell cache should be populated after creating a server.
        self.assertTrue(nova_context.CELL_CACHE)
        self.metadata.reset()
        # Cell cache should be empty after the service reset.
        self.assertEqual({}, nova_context.CELL_CACHE)

    def test_service_start_resets_cell_cache(self):
        """Tests that the cell cache for database transaction context managers
        is cleared upon a service start (example scenario: service start after
        a SIGTERM and the parent process forks child process workers).
        """
        server_req = self._build_server()
        server = self.api.post_server({'server': server_req})
        self._wait_for_state_change(server, 'ACTIVE')
        # Cell cache should be populated after creating a server.
        self.assertTrue(nova_context.CELL_CACHE)
        self.scheduler.stop()
        # NOTE(melwitt): Simulate a service starting after being stopped. The
        # scenario we want to handle is one where during start, the parent
        # process forks child process workers while one or more of its cached
        # database transaction context managers is inside a locked code
        # section. If the child processes are forked while the lock is locked,
        # the child processes begin with an already locked lock that can never
        # be acquired again. The result is that requests gets stuck and fail
        # with a CellTimeout error.
        self.scheduler.start()
        # Cell cache should be empty after the service start.
        self.assertEqual({}, nova_context.CELL_CACHE)

        # Now test the WSGI service.
        server = self.api.post_server({'server': server_req})
        self._wait_for_state_change(server, 'ACTIVE')
        # Cell cache should be populated after creating a server.
        self.assertTrue(nova_context.CELL_CACHE)
        # we need to mock nova.utils.raise_if_old_compute() that is run at
        # service startup as that will check the global service level which
        # populates the cell cache
        with mock.patch("nova.utils.raise_if_old_compute"):
            self.metadata.stop()
            self.metadata.start()
        # Cell cache should be empty after the service reset.
        self.assertEqual({}, nova_context.CELL_CACHE)


class TestOldComputeCheck(
        test.TestCase, integrated_helpers.InstanceHelperMixin):

    def test_conductor_fails_to_start_with_old_compute(self):
        old_version = service.SERVICE_VERSION_ALIASES[
            service.OLDEST_SUPPORTED_SERVICE_VERSION] - 1
        with mock.patch(
                "nova.objects.service.get_minimum_version_all_cells",
                return_value=old_version):
            self.assertRaises(
                exception.TooOldComputeService, self.start_service,
                'conductor')

    def test_api_fails_to_start_with_old_compute(self):
        old_version = service.SERVICE_VERSION_ALIASES[
            service.OLDEST_SUPPORTED_SERVICE_VERSION] - 1
        with mock.patch(
                "nova.objects.service.get_minimum_version_all_cells",
                return_value=old_version):
            self.assertRaises(
                exception.TooOldComputeService, self.useFixture,
                nova_fixtures.OSAPIFixture(api_version='v2.1'))

    def test_compute_fails_to_start_with_old_compute(self):
        old_version = service.SERVICE_VERSION_ALIASES[
            service.OLDEST_SUPPORTED_SERVICE_VERSION] - 1
        with mock.patch(
                "nova.objects.service.get_minimum_version_all_cells",
                return_value=old_version):
            self.assertRaises(
                exception.TooOldComputeService, self._start_compute, 'host1')



# Copyright 2015 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from oslo_utils import importutils

from nova import exception
from nova.objects import fields

libosinfo = None
LOG = logging.getLogger(__name__)

# TODO(vladikr) The current implementation will serve only as a temporary
# solution, due to it's dependency on the libosinfo gobject library.
# In the future it will be replaced by a pure python library or by a direct
# parsing of the libosinfo XML files. However, it will be possible only when
# libosinfo project will declare the XML structure to be a stable ABI.


class _OsInfoDatabase(object):

    _instance = None

    def __init__(self):

        global libosinfo
        try:
            if libosinfo is None:
                libosinfo = importutils.import_module(
                                             'gi.repository.Libosinfo')
        except ImportError as exp:
            LOG.info("Cannot load Libosinfo: (%s)", exp)
        else:
            self.loader = libosinfo.Loader()
            self.loader.process_default_path()

            self.db = self.loader.get_db()
            self.oslist = self.db.get_os_list()

    @classmethod
    def get_instance(cls):
        """Get libosinfo connection
        """
        if cls._instance is None:
            cls._instance = _OsInfoDatabase()

        return cls._instance

    def get_os(self, os_name):
        """Retrieve OS object based on id, unique URI identifier of the OS
           :param os_name: id - the unique operating systemidentifier
                           e.g. http://fedoraproject.org/fedora/21,
                           http://microsoft.com/win/xp,
                           or a
                           short-id - the short name of the OS
                           e.g. fedora21, winxp
           :returns: The operation system object Libosinfo.Os
           :raise exception.OsInfoNotFound: If os hasn't been found
        """
        if libosinfo is None:
            return
        if not os_name:
            raise exception.OsInfoNotFound(os_name='Empty')
        fltr = libosinfo.Filter.new()
        flt_field = 'id' if os_name.startswith('http') else 'short-id'
        fltr.add_constraint(flt_field, os_name)
        filtered = self.oslist.new_filtered(fltr)
        list_len = filtered.get_length()
        if not list_len:
            raise exception.OsInfoNotFound(os_name=os_name)
        return filtered.get_nth(0)


class OsInfo(object):
    """OS Information Structure
    """

    def __init__(self, os_name):
        self._os_obj = self._get_os_obj(os_name)

    def _get_os_obj(self, os_name):
        if os_name is not None:
            try:
                return _OsInfoDatabase.get_instance().get_os(os_name)
            except exception.NovaException as e:
                LOG.warning("Cannot find OS information - Reason: (%s)", e)

    @property
    def network_model(self):
        if self._os_obj is not None:
            fltr = libosinfo.Filter()
            fltr.add_constraint("class", "net")
            devs = self._os_obj.get_all_devices(fltr)
            if devs.get_length():
                net_model = devs.get_nth(0).get_name()
                # convert to valid libvirt values
                if net_model in ['virtio-net', 'virtio1.0-net']:
                    return 'virtio'
                # ignore any invalid ones
                if net_model in fields.VIFModel.ALL:
                    return net_model

    @property
    def disk_model(self):
        if self._os_obj is not None:
            fltr = libosinfo.Filter()
            fltr.add_constraint("class", "block")
            devs = self._os_obj.get_all_devices(fltr)
            if devs.get_length():
                disk_model = devs.get_nth(0).get_name()
                # convert to valid libvirt values
                if disk_model in ['virtio-block', 'virtio1.0-block']:
                    return 'virtio'
                # ignore any invalid ones
                if disk_model in fields.DiskBus.ALL:
                    return disk_model


class HardwareProperties(object):

    def __init__(self, image_meta):
        """:param image_meta:  ImageMeta object
        """
        self.img_props = image_meta.properties
        os_key = self.img_props.get('os_distro')
        self.os_info_obj = OsInfo(os_key)

    @property
    def network_model(self):
        return self.img_props.get('hw_vif_model',
                                  self.os_info_obj.network_model)

    @property
    def disk_model(self):
        return self.img_props.get('hw_disk_bus',
                                  self.os_info_obj.disk_model)
